make build_c && \
export GI_TYPELIB_PATH="${PWD}/build/subprojects/libinsane-gobject/src" && \
export LD_LIBRARY_PATH="${PWD}/build/subprojects/libinsane-gobject/src:build/subprojects/libinsane/src" && \
export VIRTUAL_ENV="libinsane"
