#ifndef __LIBINSANE_GOBJECT_ITEM_TYPE_H
#define __LIBINSANE_GOBJECT_ITEM_TYPE_H

typedef enum
{
	LIBINSANE_ITEM_TYPE_DEVICE,
	LIBINSANE_ITEM_TYPE_FLATBED,
	LIBINSANE_ITEM_TYPE_ADF,
	LIBINSANE_ITEM_TYPE_UNIDENTIFIED = -1,
} LibinsaneItemType;

#endif
