#ifndef __LIBINSANE_GOBJECT_IMG_FORMAT_H
#define __LIBINSANE_GOBJECT_IMG_FORMAT_H

typedef enum
{
	LIBINSANE_IMG_FORMAT_RAW_RGB_24,

	LIBINSANE_IMG_FORMAT_GRAYSCALE_8,
	LIBINSANE_IMG_FORMAT_BW_1,
	LIBINSANE_IMG_FORMAT_BMP,
	LIBINSANE_IMG_FORMAT_GIF,
	LIBINSANE_IMG_FORMAT_JPEG,
	LIBINSANE_IMG_FORMAT_PNG,
	LIBINSANE_IMG_FORMAT_CIFF,
	LIBINSANE_IMG_FORMAT_EXIF,
	LIBINSANE_IMG_FORMAT_FLASHPIX,
	LIBINSANE_IMG_FORMAT_ICO,
	LIBINSANE_IMG_FORMAT_JPEG2K,
	LIBINSANE_IMG_FORMAT_JPEG2KX,
	LIBINSANE_IMG_FORMAT_MEMORYBMP,
	LIBINSANE_IMG_FORMAT_PHOTOCD,
	LIBINSANE_IMG_FORMAT_PICT,
	LIBINSANE_IMG_FORMAT_TIFF,
} LibinsaneImgFormat;

#endif
