#ifndef __LIBINSANE_NORMALIZERS_RAW24_H
#define __LIBINSANE_NORMALIZERS_RAW24_H

void unpack_8_to_24(void *buffer, size_t *buffer_size);
void unpack_1_to_24(void *buffer, size_t *buffer_size);

#endif
